module dfc.direct;

import std.string;

import win32.directx.directdraw;

import dfc.core;
import dfc.bitmap;

// D foundation Classes
//
// DirectX windowed or full screen app
//

uint getSurfaceDeviceColour( IDirectDrawSurface surface, COLORREF rgb )
{
	HDC hdc = null;
	COLORREF tmp_rgb;
	HRESULT result;
	DDSURFACEDESC descriptor;
	uint colour;

	if ( surface.GetDC( &hdc ) == DD_OK )
	{
		tmp_rgb = GetPixel( hdc, 0, 0 );
		SetPixel( hdc, 0, 0, rgb );
		surface.ReleaseDC( hdc );
	}

	descriptor.dwSize = DDSURFACEDESC.size;

	while( DD_OK != (result = surface.Lock(	
									null, 
									&descriptor, 
									DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT | DDLOCK_READONLY, 
									null )) ) 
	{
		if ( result == DDERR_SURFACELOST )
		{
			surface.Restore();
			continue;
		}
		if ( result == DDERR_WASSTILLDRAWING )
		{
			continue;
		}
		return 0;
	}
	colour = *(cast(uint*)descriptor.lpSurface);
	colour &= ((1 << descriptor.ddpfPixelFormat.dwRGBBitCount) - 1);

	surface.Unlock( descriptor.lpSurface );

	if ( surface.GetDC( &hdc ) == DD_OK )
	{
		SetPixel( hdc, 0, 0, tmp_rgb );
		surface.ReleaseDC( hdc );
	}
	return colour;
}

void setSurfaceSrcKey( IDirectDrawSurface surface, COLORREF rgb_src_key )
{
	DDCOLORKEY key;
	key.dwColorSpaceLowValue = key.dwColorSpaceHighValue = getSurfaceDeviceColour( surface, rgb_src_key );
	surface.SetColorKey( DDCKEY_SRCBLT, &key );
}

class DirectApp : DApp
{
	static HMODULE ddModule = null; /*!< handle to ddraw.dll (if loaded) */
	static IDirectDraw directDraw = null; /*!< com interface to direct draw */

	/* possible screen modes most are not supported under win32 */
	enum ScreenColourDepth {
		SCD_DEFAULT,      /* use what ever the system is running */
		SCD_8BIT_INDEXED, /* 256 colours indexed palette */
		SCD_16BIT_565RGB, /* 16 bit vga 565  rrrrrggggggbbbbb */
		SCD_16BIT_555BGR, /* 15 bit gba x555 xbbbbbgggggrrrrr */
		SCD_24BIT_RGB,    /* 24 bit  888   RGB */
		SCD_24BIT_BGR,    /* 24 bit  888   BGR */
		SCD_32BIT_XRGB,   /* 32 bit 8888  xRGB */
		SCD_32BIT_XBGR,   /* 32 bit 8888  xBGR */
		SCD_32BIT_RGBX,   /* 32 bit 8888  RGBx (xBGR) on opposite endian systems */
		SCD_32BIT_BGRX,   /* 32 bit 8888  BGRx (xRGB) on opposite endian systems */
	}
	const ScreenColourDepth defaultFullScreen = ScreenColourDepth.SCD_16BIT_565RGB; /* SCD_32BIT_XRGB */
	
	enum ScreenDisplayMode {
		SDM_WINDOWED,
		SDM_WINDOWED_NORESIZE,
		SDM_WINDOWED_NOTITLE,
		SDM_WINDOWED_NOFRAME,
		SDM_320x200,
		SDM_320x240,
		SDM_640x480,
		SDM_800x600,
		SDM_1024x768,
		SDM_1280x1024,
		SDM_1600x1200
	}

	static IDirectDraw getDirectDraw() { return directDraw; }

	IDirectDrawSurface primSurface = null;
	IDirectDrawSurface backingSurface = null;
	IDirectDrawClipper clipper = null;
	/* it seems you need the primary pixel format to make an offscreen buffer */
	DDSURFACEDESC descriptor;

	char[] title;
	public int backingWidth;
	public int backingHeight;

	public this( char[] title0 )
	{
		super();
		title = title0;
	}

	public this()
	{
		this( "" );
	}

	/* these modes are always supported */
	public ScreenColourDepth[] getPossiblyColourDepth()
	{
		ScreenColourDepth[] modes;
		modes ~= ScreenColourDepth.SCD_DEFAULT;
		modes ~= ScreenColourDepth.SCD_16BIT_565RGB;
		modes ~= ScreenColourDepth.SCD_32BIT_XRGB;
		return modes;
	}

	private void getWidthHeightFor( inout int width, inout int height, ScreenDisplayMode mode )
	{
		switch( mode )
		{
		case ScreenDisplayMode.SDM_WINDOWED:
		case ScreenDisplayMode.SDM_WINDOWED_NORESIZE:
		case ScreenDisplayMode.SDM_WINDOWED_NOTITLE:
		case ScreenDisplayMode.SDM_WINDOWED_NOFRAME:
			break;
		case ScreenDisplayMode.SDM_320x200:
			width = 320; height = 200;
			break;
		case ScreenDisplayMode.SDM_320x240:
			width = 320; height = 240;
			break;
		case ScreenDisplayMode.SDM_640x480:
			width = 640; height = 480;
			break;
		case ScreenDisplayMode.SDM_800x600:
			width = 800; height = 600;
			break;
		case ScreenDisplayMode.SDM_1024x768:
			width = 1024; height = 768;
			break;
		case ScreenDisplayMode.SDM_1280x1024:
			width = 1280; height = 1024;
			break;
		case ScreenDisplayMode.SDM_1600x1200:
			width = 1600; height = 1200;
			break;
		default:
			throw new Exception("unknown Screen display mode.");
		}
	}

	private bit createWindow( int width, int height, ScreenDisplayMode mode, out uint dxMode )
	{
		dxMode = DDSCL_NORMAL;
		switch( mode )
		{
		case ScreenDisplayMode.SDM_WINDOWED:
			create(	WS_THICKFRAME | WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_SYSMENU | WS_VISIBLE, 
						0,
						title, 
						width, height );
			setClientSize( width, height );
			break;
		case ScreenDisplayMode.SDM_WINDOWED_NORESIZE:
			create(	WS_DLGFRAME | WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_SYSMENU | WS_VISIBLE, 
						0,
						title, 
						width, height );
			setClientSize( width, height );
			break;
		case ScreenDisplayMode.SDM_WINDOWED_NOTITLE:
			create(	(WS_DLGFRAME | WS_VISIBLE | WS_POPUP), 
						WS_EX_WINDOWEDGE,
						title, 
						width, height );
			setClientSize( width, height );
			break;
		case ScreenDisplayMode.SDM_WINDOWED_NOFRAME:
			create(	WS_VISIBLE|WS_POPUP, 
						WS_EX_WINDOWEDGE,
						title, 
						width, height );
			setClientSize( width, height );
			break;
		default:
			dxMode = DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN;
			create(	WS_VISIBLE|WS_POPUP, 
						0,
						title, 
						width, height );
			break;
		}
		return (getHwnd() != null);
	}

	private bit createSurfaces( int width, int height, int bits, uint dxMode )
	{
		DDPIXELFORMAT format;
		backingWidth  = width;
		backingHeight = height;

		if ( DD_OK != directDraw.SetCooperativeLevel( getHwnd(), dxMode ) )
		{
			return false;
		}
		if ( (dxMode & DDSCL_FULLSCREEN) != 0 )
		{
			if ( DD_OK != directDraw.SetDisplayMode( width, height, bits ) )
			{
				return false;
			}
		}

		memset( &descriptor, 0, DDSURFACEDESC.size );

		descriptor.dwSize  = DDSURFACEDESC.size;
		descriptor.dwFlags = DDSD_CAPS;
		descriptor.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_VIDEOMEMORY;
	
		if ( DD_OK != directDraw.CreateSurface( &descriptor, &primSurface, null ) )
		{
			return false;
		}

		descriptor.dwSize  = DDSURFACEDESC.size;
		descriptor.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		descriptor.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		descriptor.dwWidth  = width;
		descriptor.dwHeight = height;

		if ( DD_OK != directDraw.CreateSurface( &descriptor, &backingSurface, null ) )
		{
			return false;
		}

		if ( DD_OK != directDraw.CreateClipper( 0, &clipper, null ) )
		{
			return false;
		}

		if ( DD_OK != clipper.SetHWnd( 0, getHwnd() ) )
		{
			return false;
		}

		if ( DD_OK != primSurface.SetClipper( clipper ) )
		{
			return false;
		}

		memset( &format, 0, DDPIXELFORMAT.size );
		format.dwSize = DDPIXELFORMAT.size;

		if ( DD_OK != primSurface.GetPixelFormat( &format ) )
		{
			return false;
		}

		if ( (format.dwFlags & DDPF_RGB) == 0 )
		{
			return false;
		}

		if ( bits == 0 )
		{
			return true;
		}
		else if ( bits == 16 )
		{
			if (		(format.dwRBitMask != 0x0000F800) 
					||	(format.dwGBitMask != 0x000007E0)
					||	(format.dwBBitMask != 0x0000001F) )
			{
				return false;
			}
		}
		else if ( bits == 32 )
		{
			if (		(format.dwRBitMask != 0x00FF0000) 
					||	(format.dwGBitMask != 0x0000FF00)
					||	(format.dwBBitMask != 0x000000FF) )
			{
				return false;
			}
		}
		else
		{
			return false;
		}

		return true;
	}

	IDirectDrawSurface offScreen = null;


	IDirectDrawSurface createOffscreen( int w, int h )
	{
		DDSURFACEDESC desc;

		memset( &desc, 0, DDSURFACEDESC.size );
		desc = descriptor;

		desc.dwSize  = DDSURFACEDESC.size;
		desc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		desc.dwWidth  = cast(DWORD)w;
		desc.dwHeight = cast(DWORD)h;

		printf( "w:%d h:%d desc.w:%d desc.h:%d", w, h, desc.dwWidth, desc.dwHeight );

		if ( DD_OK == directDraw.CreateSurface( &desc, &offScreen, null ) )
		{
			printf("888 made a surface;\n");
			return offScreen;
		}
		else
		{
			printf("888 can not make a surface;\n");
			return null;
		}
	}

	private bit initDefault( int width, int height, ScreenDisplayMode mode )
	{
		uint dxMode;
		if ( !createWindow( width, height, mode, dxMode ) )
		{
			return false;
		}
		return createSurfaces( width, height, 0, dxMode );
	}

	private bit init565( int width, int height, ScreenDisplayMode mode )
	{
		uint dxMode;
		if ( !createWindow( width, height, mode, dxMode ) )
		{
			return false;
		}
		return createSurfaces( width, height, 16, dxMode );
	}

	private bit initRGB32( int width, int height, ScreenDisplayMode mode )
	{
		uint dxMode;
		if ( !createWindow( width, height, mode, dxMode ) )
		{
			return false;
		}
		return createSurfaces( width, height, 32, dxMode );
	}

	/*! only in SDM_WINDOWED_*** are the width, height params used */
	public bit initialise( int width, int height, ScreenColourDepth colourDepth, ScreenDisplayMode mode )
	{
		if ( !loadDirectXLibs() )
		{
			MessageBox( null, "unable to load dx libs", "DxOk", MB_OK );
			return false;
		}
		getWidthHeightFor( width, height, mode );

		switch( colourDepth )
		{
		case ScreenColourDepth.SCD_DEFAULT:
			if ( !initDefault( width, height, mode ) )
			{
				return false;
			}
			break;

		case ScreenColourDepth.SCD_16BIT_565RGB:
			if ( !init565( width, height, mode ) )
			{
				return false;
			}
			break;

		case ScreenColourDepth.SCD_32BIT_XRGB:
			if ( !initRGB32( width, height, mode ) )
			{
				return false;
			}
			break;

		case ScreenColourDepth.SCD_8BIT_INDEXED:
		case ScreenColourDepth.SCD_16BIT_555BGR:
		case ScreenColourDepth.SCD_24BIT_RGB:
		case ScreenColourDepth.SCD_24BIT_BGR:
		case ScreenColourDepth.SCD_32BIT_XBGR:
		case ScreenColourDepth.SCD_32BIT_RGBX:
		case ScreenColourDepth.SCD_32BIT_BGRX:
		default:
			throw new Exception( "Unsupported screen mode" );
			return false;
		}


		showWindow( true );
		centerWindow();
		return true;
	}

	public IDirectDrawSurface getBackSurface() 
	{ 
		return backingSurface; 
	}

	alias void (*updateFunc)( inout DDSURFACEDESC desc, void * screen, int scanWidth );
	alias void delegate( inout DDSURFACEDESC desc, void * screen, int scanWidth ) updateDelegateFunc;

	class updateWrapperObj
	{
		updateFunc toCall;
		this( updateFunc toCall0 ) { toCall = toCall0; }
		
		void callFunc( inout DDSURFACEDESC desc, void * screen, int scanWidth )
		{
			toCall( desc, screen, scanWidth );
		}

		static updateDelegateFunc getDelegateFor( updateFunc funcPtr )
		{
			updateWrapperObj obj;
			obj = new updateWrapperObj( funcPtr );
			return &obj.callFunc;
		}
	}

	public void updateFrame( updateFunc funcPtr )
	{
		updateFrame( updateWrapperObj.getDelegateFor( funcPtr ) );
	}

	public void updateFrame( updateDelegateFunc delegatePtr )
	{
		int x, y, i, tmp;
		DDSURFACEDESC descriptor;
		HRESULT result;

		descriptor.dwSize = DDSURFACEDESC.size;
	
		while( DD_OK != (result = backingSurface.Lock(	
										null, 
										&descriptor, 
										DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT | DDLOCK_WRITEONLY, 
										null )) ) 
		{
			if ( result == DDERR_SURFACELOST )
			{
				primSurface.Restore();
				backingSurface.Restore();
				continue;
			}
			return;
		}
		delegatePtr( descriptor, descriptor.lpSurface, descriptor.lPitch );

		backingSurface.Unlock( descriptor.lpSurface );
	}

	public void updatePrimarySurface() 
	{
		RECT source;
		RECT destination;
		POINT point;
		int h, w;

		// get origin of client area
		point.x = 0;
		point.y = 0;
		ClientToScreen( hwnd, &point );

		// get window client area
		GetClientRect( hwnd, &destination );
		w = backingWidth;
		h = backingHeight;
		if ( (destination.right - destination.left) < w )
		{
			w = (destination.right - destination.left);
		}
		if ( (destination.bottom - destination.top) < h )
		{
			h = (destination.bottom - destination.top);
		}

		// offset destination rectangle
		destination.left   = point.x;
		destination.top    = point.y;
		destination.right  = point.x + w;
		destination.bottom = point.y + h;

		// setup source rectangle
		source.left   = 0;
		source.top    = 0;
		source.right  = w;
		source.bottom = h;
/*
		printf( "blt [(%d,%d)-(%d,%d)]-> [(%d,%d)-(%d,%d)]\n",
			source.left,
			source.top,
			source.right,
			source.bottom,
			destination.left,
			destination.top,
			destination.right,
			destination.bottom
			);
	*/
		// blt secondary to primary surface
		primSurface.Blt( &destination, backingSurface, &source, DDBLT_WAIT, null );
	}

	public bit waitForNextFrame()
	{
		if ( DFC.checkForMsgs() ) { return true; }
		directDraw.WaitForVerticalBlank( DDWAITVB_BLOCKBEGIN, null );
		return false;
	}


	public bit processNextFrame( updateDelegateFunc delegatePtr )
	{
		if ( DFC.checkForMsgs() ) { return true; }
		directDraw.WaitForVerticalBlank( DDWAITVB_BLOCKBEGIN, null );
		updateFrame( delegatePtr );
		updatePrimarySurface();
		return false;
	}

	public bit processNextFrame( updateFunc funcPtr )
	{
		return processNextFrame( updateWrapperObj.getDelegateFor( funcPtr ) );
	}

	static void freeDirectXLibs()
	{
		if ( ddModule != null ) // already loaded
		{
			FreeLibrary( ddModule );
			ddModule = null;
		}
	}

	HDC getBackDC()
	{
		HDC hdc = null;
		backingSurface.GetDC( &hdc );
		return hdc;
	}

	void releaseBackDC( HDC hdc )
	{
		backingSurface.ReleaseDC( hdc );
	}

	uint getDeviceColour( COLORREF rgb )
	{
		return getSurfaceDeviceColour( backingSurface, rgb );
	}

	void setSrcKey( COLORREF rgb_src_key )
	{
		DDCOLORKEY key;
		key.dwColorSpaceLowValue = key.dwColorSpaceHighValue = getDeviceColour( rgb_src_key );
		backingSurface.SetColorKey( DDCKEY_SRCBLT, &key );
	}

	void renderBitmap( DirectBitmap bitmap, int x, int y )
	{
		bitmap.directRenderOnto( backingSurface, x, y );
	}

	static bit loadDirectXLibs()
	{
		if ( ddModule === null ) // already loaded
		{
			ddModule = LoadLibraryA( "ddraw.dll" );
			if ( ddModule === null ) { return false; }
			ddCreateFunc createFuncPtr;
			createFuncPtr = cast(ddCreateFunc)GetProcAddress( ddModule, "DirectDrawCreate" );
			if ( createFuncPtr == null )
			{
				freeDirectXLibs();
				return false;
			}
			directDraw = null;
			if ( DD_OK != createFuncPtr( null, &directDraw, null ) ) 
			{
				freeDirectXLibs();
				return false;
			}
			if( directDraw === null )
			{
				freeDirectXLibs();
				return false;
			}
		}
		return true;
	}
}

class DirectBitmap : DBitmap
{
	IDirectDrawSurface offScreen = null;

	this() { super(); }

	this( DirectApp ddApp, char[] name ) 
	{ 
		HDC hdc;
		HRESULT rv;

		super( name ); 

		if ( (offScreen = ddApp.createOffscreen( getWidth(), getHeight() )) !== null )
		{
			if ( offScreen.GetDC( &hdc ) == DD_OK )
			{
				renderOnto( hdc, 0, 0 );
				offScreen.ReleaseDC( hdc );
				setSurfaceSrcKey( offScreen, 0xFF00FF );
				return;
			}
			return;
		}
	}

	void setSrcKey( COLORREF rgb_src_key )
	{
		setSurfaceSrcKey( offScreen, rgb_src_key );
	}
	
	void directRenderOnto( IDirectDrawSurface surface, int x, int y )
	{
		DDBLTFX fx;
		DWORD flags;
		RECT srcRect;
		RECT destRect;
		HRESULT rv;

		srcRect.left   = 0;
		srcRect.top    = 0;
		srcRect.right  = getWidth();
		srcRect.bottom = getHeight();

		destRect.left   = x;
		destRect.top    = y;
		destRect.right  = x + srcRect.right;
		destRect.bottom = y + srcRect.bottom;

		flags = DDBLT_KEYSRC|DDBLT_WAIT;
		memset( &fx, 0, DDBLTFX.size );
		fx.dwSize = DDBLTFX.size;

		if ( offScreen !== null )
		{
			surface.Blt( &destRect, offScreen, &srcRect, flags, null );
		}
	}
}
